package Lingua::LanguageGuesser;

use 5.006000;
use strict;
use warnings;
use Carp qw(croak);
use bytes;
use Encode::Guess;

require Exporter;
our @ISA = qw(Exporter);

our @EXPORT_OK = qw(
    create_model
    delete_model
    list_my_models
);

our $VERSION = '0.00_14'; # Now beta version
our $MaxLank = 400;      # "-t" option of origianl "TextCat"

=pod

=head1 NAME

Lingua::LanguageGuesser - Language guesser

=head1 SYNOPSIS

  use Lingua::LanguageGuesser;

  $textstring = 'text strings';
  @lang_list_sorted_similarity =
      Lingua::LanguageGuesser
           ->guess($textstring)
           ->eliminate()
           ->suspect('english', 'japanese-euc_jp')
           ->result_list();
  print "Input is perhaps $lang_list_sorted_similarity[0]";

=head1 DESCRIPTION

Lingua::LanguageGuesser guess the language a given string or file is written in.

This modlule is imprimetain of "TextCat"  Language guesser as perl module with addtional utf8 encoded text supports.

The feature is as follows.

=over 4

=item * This module is very portable, because it is writen by pure Perl program codes and not necessary to use another optional perl modules without original Perl kit.

=item * Over 60 Languaage model supported.

=item * utf8 and not-utf8 language encode devides by using Encode::Guess perl module.

=item * This module supports to create your own language model (profile) easy.

=item * This module have Object Oriented style interface for guess languages from text strings.

=back

=head1 METHODS

guess() method is constractor of the Lingua::LanguageGuesser object. 
The first parameter (hash vale) is setting options (not nessary). 'utf8' option decide treetment of "utf8 encoded mode". If text strings is guessed as utf8 encode by Encode::Guess moldule, "auto" parameter changes mode to use uft8 language model from not-utf8. "omit" parameter omit utf8 language model. "include" parameter sets language model both not-utf8 and utf8. 
If you have created your own language model by create_model() method, 'MyModel' indicates in which directory your own language models are. 'MyModel_utf8' do your own utf8 language models.
'MaxLine' indicates how many lines onyl read from the top line of text. 
The second parameter is text string for guessing languege. You can call new() method insted of guess(). The new() method is alias of The guess().

  use Lingua::LanguageGuesser;
  my $guesser = Lingua::LanguageGuesser->guess(
                    {
                      utf8         => 'include', # select "auto","omit" or "inculde".
                      MyModel      => './my_language_model_directory',
                      MyModel_utf8 =>  './my_utf8_kanguage_model_directory',
                      MaxLine'      => 1000
                    },
                    $textstring
                );

or

  my $guesser = Lingua::LanguageGuesser->new(
                    {
                      utf8         => 'include', # select "auto", "omit" or "inculde"
                      MyModel      => './my_language_model_directory',
                      MyModel_utf8 => './my_utf8_kanguage_model_directory',
                      MaxLine      => 1000
                    },
                    $textstring
                );

=cut

*guess = \&new;

# package costractor
sub new {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my %params = ();
    if (ref($_[0]) eq 'HASH') { %params = (%params, %{+shift}) }
    my $input = shift @_;
    my $self = {};
    bless ($self, $class);
    
    if (defined $params{'utf8'} and 
        ($params{'utf8'} eq 'omit' or $params{'utf8'} eq 'include')) {
    	$self->{'utf8'} = $params{'utf8'}
    }
    else {
        $self->{'utf8'} = 'auto'
    }
    $self->{'MyModel'}        = $params{'MyModel'};
    $self->{'MyModel_utf8'}   = $params{'MyModel_utf8'};
    $self->{'MaxLine'}        = $params{'MaxLine'};
    $self->{'Results'}        = {};
    $self->{'eliminate_mode'} = 0;
    $self->{'set_text'}       = 0;
    $self->set_text($input) if $input && !(ref($input));

    return $self;
}

=pod

You can set text string for guess lanaugage as follows.

  my $guesser = Lingua::LanguageGuesser->new(
                    {
                      utf8         => 'include', # select "auto", "omit" or "inculde"
                      MyModel      => './my_language_model_directory',
                      MyModel_utf8 => './my_utf8_kanguage_model_directory',
                      MaxLine      => 1000
                    }
                );
  print "$textstring1 is ", $guesser->set_text($textstring1)->best_scoring(), "\n";
  print "$textstring2 is ", $guesser->set_text($textstring2)->best_scoring(), "\n";

=cut

# Initialization of object
sub set_text {
    my ($self, $input) = @_;
    my %results = ();
    my $LM = $INC{"Lingua/LanguageGuesser.pm"};
    $LM =~ s/LanguageGuesser.pm$/LM/;
    my $LM_utf8 = $LM . '_utf8';
    croak "Set text strings for guess language !"
        unless $input or ref($input);

    $self->{'eliminate_mode'} = 0;

    if (defined $self->{'MaxLine'}) {
        my $work = ""; my $i = 0;
        foreach my $line (split /\n/, $input) {
            $work .= $line;
            $i++;
            last if $self->{'MaxLine'} <= $i;
        }
        $input = $work;
    }

    my @languages       = _get_language_model($LM);
    my @languages_utf8  = _get_language_model($LM_utf8);
    my @mylanguages     = _get_language_model($self->{'MyModel'})
        if defined $self->{'MyModel'};
    my @mylanguages_utf8 = _get_language_model($self->{'MyModel_utf8'})
        if defined $self->{'MyModel_utf8'}; 

    # guess by Out-Of-Place method
    if ($self->{'utf8'} eq 'auto') {
        my $enc = guess_encoding($input);
        if (ref($enc) and $enc->name eq 'utf8') {
            _OutOfPlace(\$input, @languages_utf8, $LM_utf8, \%results);
            _OutOfPlace(\$input, @mylanguages_utf8, $self->{'MyModel_utf8'}, \%results)
                if scalar @mylanguages_utf8;
        }
        else {
            _OutOfPlace(\$input, @languages, $LM, \%results);
            _OutOfPlace(\$input, @mylanguages, $self->{'MyModel'}, \%results)
                if scalar @mylanguages;
        }
    }
    elsif ($self->{'utf8'} eq 'include') { 
        _OutOfPlace(\$input, @languages, $LM, \%results);
        _OutOfPlace(\$input, @languages_utf8, $LM_utf8, \%results);
        _OutOfPlace(\$input, @mylanguages, $self->{'MyModel'}, \%results)
            if scalar @mylanguages;
        _OutOfPlace(\$input, @mylanguages_utf8, $self->{'MyModel_utf8'}, \%results)
            if scalar @mylanguages_utf8;
    }
    elsif ($self->{'utf8'} eq 'omit') {
        _OutOfPlace(\$input, @languages, $LM, \%results);
        _OutOfPlace(\$input, @mylanguages, $self->{'MyModel'}, \%results)
            if scalar @mylanguages;
    }
    else {}
    $self->{'set_text'} = 1;

    $self->{'Results'} = \%results;
}

=pod

Support three output methods as follows.

1. score_of_lang() method return languages with score of similariry. (perl hash value)

  %score_of_lang = $guesser->score_of_lang();
  $score_of_english = $score_of_lang{'english'};

=cut

sub score_of_lang {
    my $self = shift;
    croak "Set text strings for guess language !" unless $self->{'set_text'};
    my $results = $self->{'Results'};
    return %{ $self->{'Results'} };
}

=pod

2. result_list() method return language list sorted by similarity between the text string and languages. (perl array value)

  @lang_list_sorted_similarity = $guesser->result_list();

=cut

sub result_list {
    my $self = shift;
    croak "Set text strings for guess language !" unless $self->{'set_text'};
    my $results = $self->{'Results'};
    my @langs = (keys %$results);
    if (scalar @langs >= 2) {
        @langs = sort { $results->{$a} <=> $results->{$b} }
             keys %$results;
    }
    return @langs;
}

=pod

3. best_scoring() method retruns name of the most similar language only. (perl scalar value)
If you use eliminate() method and the result has some candidates, it returns "Two or more suspects remain" as text string. 
If you call suspect() methods and the result has no candidates, it returens "No language suspects" as text string.

  $most_similar_lang = $guesser->best_scoring()

or

  $most_similar_lang = $guesser->eliminate()->best_scoring();
  if ($most_similar_lang eq 'Two or more suspects remain') { Any Code }

=cut

sub best_scoring {
    my $self = shift;
    croak "Set text strings for guess language !" unless $self->{'set_text'};
    my $results = $self->{'Results'};
    my $best = "";
    my @langs = (keys %$results);
    my $count = scalar @langs;
    if ($count >= 2 and $self->{'eliminate_mode'}) {
        $best = "Two or more suspects remain"
    }
    elsif ($count == 0) {
        $best = "No language suspects";
    }
    elsif ($count == 1) {
        $best = shift @langs;
    }
    else {
        @langs = sort { $results->{$a} <=> $results->{$b} }
            keys %$results;
        $best = shift @langs;
    }
    return $best;
}

=pod

This module supports two methods for filtering candidate of language. eliminate() and suspect(). Order of the methods infuluences the result. Because "the first method works first" in this module. 
eliminate() method determines how much worse result must be in order not to be mentioned as an alternative. Typical value: 1.05 or 1.1.  Default value is 1.05.

  $guesser->eliminate( 1.1 ) ;

=cut

sub eliminate {
    my $self = shift;
    my $limit_ratio = shift || 1.05; # $opt_u
    croak "Set text strings for guess language !" unless $self->{'set_text'};
    my $results = $self->{'Results'};
    my @langs = sort { $results->{$a} <=> $results->{$b} } keys %$results;
    if (scalar @langs > 0) {
        my $limit = $results->{$langs[0]} * $limit_ratio;
        map { delete $results->{$_} if $results->{$_} > $limit } keys %$results;
    }
    $self->{'eliminate_mode'} = 1;
    return $self;
}

=pod

"suspect" method suspects candidates of language form the parameter list only.

  $guesser->suspect('english', 'japanese-euc_jp');

=cut

sub suspect {
    my $self = shift;
    my @languages = @_;
    croak "Set text strings for guess language !" unless $self->{'set_text'};
    my $results = $self->{'Results'};
    my @langs = sort { $results->{$a} <=> $results->{$b} } keys %$results;
    my %new_results = ();
    map { $new_results{$_} = $results->{$_} if defined $results->{$_}}
        @languages;
    $self->{'Results'} = \%new_results;
    return $self;
}

=pod

You can combine some methods as follows.

  @lang_list_sorted_similarity =
       Lingua::LanguageGuesser
           ->guess($textstring)
           ->eliminate()
           ->suspect('english', 'japanese-euc_jp')
           ->result_list();

You can create or delete your own language model (profile).
If same language name exists in default language model, your language model have priority.

create_model() method creates or remakes language model. Set your language model directory path, source file for create language model, and language name as follows.

  use Lingua::LanguageGuesse qw(create_model);
  create_model('./my_language_model_directory', 'source_file', $language_name );

=cut

sub create_model {
    my ($LM_pass, $source_file, $LM_name) = @_;
    my $input = "";
    my %ngram = ();
    open (SOURCE, "$source_file") or croak "$!\n";
    {
        local $/ = undef; 
        $input = <SOURCE>;
    }
    my @result = _create_lm(\$input, \%ngram);
    open (LM, ">${LM_pass}/${LM_name}.lm") or croak "$!\n";
    print LM join("\n", map { "$_\t $ngram{$_}"; } @result),"\n";
    close LM;
}

=pod

delete_model() method deletes language model. Set your language model directory path and language name you want to delete.

  use Lingua::LanguageGuesser qw(delete_model);
  delete_model('./my_language_model_directory', $language_name);

=cut

sub delete_model {
    my ($LM_pass, $LM_name) = @_;
    unlink "${LM_pass}/${LM_name}.lm" or croak "$!\n";
}

=pod

list_my_models() method retrun your own langugae model list.

  use Lingua::LanguageGuesser qw(list_my_model);
  my @mymodel = list_my_model('./my_language_model_directory');

=cut

# get list of language models in the directory.
sub list_my_models  {
   my $pass = shift;
   my $langs = _get_language_model($pass);
   return @$langs;
}

# Calculation similarity by Out-of-place.
sub _OutOfPlace {
    my ($input, $languages, $LM, $results) = @_;

    # create ngrams for input.
    my @unknown = _create_lm($input);

    # load model and count for each language.
    foreach my $language (@$languages) {
        # loads the language model into hash %$language.
        my %ngram = ();
        my $rang = 1;
        open(LM, "$LM/${language}.lm") || croak "cannot open ${language}.lm: $!";
        while (defined (my $line = <LM>)) {
            chomp $line;
            # only use lines starting with appropriate character. Others are
            # ignored.
            if ($line =~ /^[^0-9\s]+/o) {
                $ngram{$&} = $rang++
            }
        }
        close(LM);

        # compares the language model with input ngrams list
        my ($i, $p) = (0, 0);
        while ($i < scalar @unknown) {
            if ($ngram{$unknown[$i]}) {
                $p +=  abs($ngram{$unknown[$i]} - $i)
            }
            else {
                $p +=  $MaxLank
            }
            ++$i;
        }
        $results->{$language} = $p;
    }
}

# Create Language Model by N-gram.
sub _create_lm {
    my $text  = shift;
    my $ngram = shift;

    # then add the ngrams found in each word in the hash
    foreach my $word (split(/[0-9\s]+/, $$text)) {
        $word = "_" . $word . "_";
        my $len = length($word);
        my $flen = $len;
        for (my $i = 0; $i < $flen; $i++) {
            $ngram->{substr($word, $i, 5)}++ if $len > 4;
            $ngram->{substr($word, $i, 4)}++ if $len > 3;
            $ngram->{substr($word, $i, 3)}++ if $len > 2;
            $ngram->{substr($word, $i, 2)}++ if $len > 1;
            $ngram->{substr($word, $i, 1)}++;
            $len--;
        }
    }

    # as suggested by Karel P. de Vos, k.vos@elsevier.nl, we speed up
    # sorting by removing singletons
    #my $mimimam_of_ngram = 0;
    #map { my $key = $_; if ($ngram->{$key} <= $mimimam_of_ngram)
    #     { delete $ngram->{$key} } } keys %$ngram;
    #however I have very bad results for short inputs, this way

    # sort the ngrams, and spit out the $MaxLank frequent ones.
    # adding  `or $a cmp $b' in the sort block makes sorting five
    # times slower..., although it would be somewhat nicer (unique result)
    my @sorted = sort { $ngram->{$b} <=> $ngram->{$a} } keys %$ngram;
    splice(@sorted, $MaxLank) if scalar @sorted > $MaxLank;

    return @sorted;
}

# get list of language models
sub _get_language_model {
    my $pass = shift;
    # open directory to find which languages are supported
    opendir DIR, "$pass" or croak "directory $pass: $!\n";
    my @languagemodel = grep { s/\.lm$// && -r "${pass}/${_}.lm" } readdir(DIR);
    closedir DIR;
    return \@languagemodel;
}

1;
__END__

=head2 LANGUAGE SUPPORTED (not utf8 mode)

=over 4

=item afrikaans

=item albanian

=item amharic-utf

=item arabic-iso8859_6

=item arabic-windows1256

=item armenian

=item basque

=item belarus-windows1251

=item bosnian

=item breton

=item bulgarian-iso8859_5

=item catalan

=item chinese-big5

=item chinese-gb2312

=item croatian-ascii

=item czech-iso8859_2

=item danish

=item dutch

=item english

=item esperanto

=item estonian

=item finnish

=item french

=item frisian

=item georgian

=item german

=item greek-iso8859-7

=item hebrew-iso8859_8

=item hindi

=item hungarian

=item icelandic

=item indonesian

=item irish

=item italian

=item japanese-euc_jp

=item japanese-shift_jis

=item korean

=item latin

=item latvian

=item lithuanian

=item malay

=item manx

=item marathi

=item middle_frisian

=item mingo

=item nepali

=item norwegian

=item persian

=item polish

=item portuguese

=item quechua

=item romanian

=item rumantsch

=item russian-iso8859_5

=item russian-koi8_r

=item russian-windows1251

=item sanskrit

=item scots

=item scots_gaelic

=item serbian-ascii

=item slovak-ascii

=item slovak-windows1250

=item slovenian-ascii

=item slovenian-iso8859_2

=item spanish

=item swahili

=item swedish

=item tagalog

=item tamil

=item thai

=item turkish

=item ukrainian-koi8_u

=item vietnamese

=item welsh

=item yiddish-utf

=back

=head2 LANGUAGE SUPPORTED (utf8 mode)

=over 4

=item amharic-utf

=item basque

=item bosnian

=item chinese_simple-utf8

=item chinese_ZH-utf8

=item croatian-ascii

=item english

=item finnish-utf8

=item french-utf8

=item german-utf8

=item indonesian

=item italian-utf8

=item japanese-utf8

=item korean-utf8

=item latin

=item malay

=item manx

=item norwegian-utf8

=item romanian

=item russian-iso8859_5

=item sanskrit

=item scots

=item serbian-ascii

=item slovak-ascii

=item slovenian-ascii

=item spanish-utf8

=item swahili

=item swedish-utf8

=item tagalog

=item welsh

=item yiddish-utf

=back

=head2 EXPORT

The module exports nothing by default. create_model(), delete_model() and  list_my_models() only can be exported on demand, as in

  use Lingua::LanguageGuesser qw(create_model delete_model list_my_models);

=head1 SEE ALSO

Another Lnaguage Guesser in CPAN

  Lingua::Identify
  Language::Guess
  Text::Language::Guess
  Text::Ngram::LanguageDetermine 


"TextCat" infomation

  http://www.let.rug.nl/~vannoord/TextCat/

Our "Gensen" project Home Page

  http://gensen.dl.itc.u-tokyo.ac.jp/

=head1 AUTHOR

Akira Maeda <maeda@lib.u-tokyo.ac.jp>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2006 by Akira Maeda (maeda@lib.u-tokyo.ac.jp) 
Original Souce Code "TextCat" was written by Gertjan van Noord.

This library is free software; you can redistribute it and/or modify it under the terms of  the GNU General Public License (GPL) as published by the Free Software Foundation (http://www.fsf.org/); either version 2 of the License, or (at your option) any later version.

=cut
